@@ -3,14 +3,16 @@ package vsphere
 import (
 	"context"
 	"encoding/base64"
+	"errors"
 	"fmt"
 	"strings"
 
 	"github.com/google/uuid"
 	machinev1 "github.com/openshift/machine-api-operator/pkg/apis/machine/v1beta1"
-	vsphereapi "github.com/openshift/machine-api-operator/pkg/apis/vsphereprovider/v1alpha1"
+	vspherev1 "github.com/openshift/machine-api-operator/pkg/apis/vsphereprovider/v1beta1"
 	machinecontroller "github.com/openshift/machine-api-operator/pkg/controller/machine"
-	"github.com/pkg/errors"
+	"github.com/openshift/machine-api-operator/pkg/controller/vsphere/session"
+	"github.com/vmware/govmomi/find"
 	"github.com/vmware/govmomi/object"
 	"github.com/vmware/govmomi/property"
 	"github.com/vmware/govmomi/vapi/rest"
@@ -23,13 +25,14 @@ import (
 )
 
 const (
-	minMemMB         = 2048
-	minCPU           = 2
-	diskMoveType     = string(types.VirtualMachineRelocateDiskMoveOptionsMoveAllDiskBackingsAndConsolidate)
-	ethCardType      = "vmxnet3"
-	providerIDPrefix = "vsphere://"
-	regionKey        = "region"
-	zoneKey          = "zone"
+	minMemMB              = 2048
+	minCPU                = 2
+	fullCloneDiskMoveType = string(types.VirtualMachineRelocateDiskMoveOptionsMoveAllDiskBackingsAndConsolidate)
+	linkCloneDiskMoveType = string(types.VirtualMachineRelocateDiskMoveOptionsCreateNewChildDiskBacking)
+	ethCardType           = "vmxnet3"
+	providerIDPrefix      = "vsphere://"
+	regionKey             = "region"
+	zoneKey               = "zone"
 )
 
 // These are the guestinfo variables used by Ignition.
@@ -37,6 +40,7 @@ const (
 const (
 	GuestInfoIgnitionData     = "guestinfo.ignition.config.data"
 	GuestInfoIgnitionEncoding = "guestinfo.ignition.config.data.encoding"
+	GuestInfoHostname         = "guestinfo.hostname"
 )
 
 // Reconciler runs the logic to reconciles a machine resource towards its desired state
@@ -52,30 +56,43 @@ func newReconciler(scope *machineScope) *Reconciler {
 
 // create creates machine if it does not exists.
 func (r *Reconciler) create() error {
-	if err := validateMachine(*r.machine, *r.providerSpec); err != nil {
+	if err := validateMachine(*r.machine); err != nil {
 		return fmt.Errorf("%v: failed validating machine provider spec: %v", r.machine.GetName(), err)
 	}
 
-	moTask, err := r.session.GetTask(r.Context, r.providerStatus.TaskRef)
-	if err != nil {
-		if !isRetrieveMONotFound(r.providerStatus.TaskRef, err) {
+	if r.providerStatus.TaskRef != "" {
+		moTask, err := r.session.GetTask(r.Context, r.providerStatus.TaskRef)
+		if err != nil {
+			if !isRetrieveMONotFound(r.providerStatus.TaskRef, err) {
+				return err
+			}
+		}
+		if taskIsFinished, err := taskIsFinished(moTask); err != nil || !taskIsFinished {
+			if !taskIsFinished {
+				return fmt.Errorf("task %v has not finished", moTask.Reference().Value)
+			}
 			return err
 		}
 	}
-	if taskIsFinished, err := taskIsFinished(moTask); err != nil || !taskIsFinished {
-		if !taskIsFinished {
-			return fmt.Errorf("task %v has not finished", moTask.Reference().Value)
-		}
-		return err
-	}
 
 	if _, err := findVM(r.machineScope); err != nil {
-		if !IsNotFound(err) {
+		if !isNotFound(err) {
 			return err
 		}
 		if r.machineScope.session.IsVC() {
 			klog.Infof("%v: cloning", r.machine.GetName())
-			return clone(r.machineScope)
+			task, err := clone(r.machineScope)
+			if err != nil {
+				conditionFailed := conditionFailed()
+				conditionFailed.Message = err.Error()
+				statusError := setProviderStatus(task, conditionFailed, r.machineScope, nil)
+				if statusError != nil {
+					return fmt.Errorf("Failed to set provider status: %w", err)
+				}
+				return err
+			}
+
+			return setProviderStatus(task, conditionSuccess(), r.machineScope, nil)
 		}
 		return fmt.Errorf("%v: not connected to a vCenter", r.machine.GetName())
 	}
@@ -85,29 +102,31 @@ func (r *Reconciler) create() error {
 
 // update finds a vm and reconciles the machine resource status against it.
 func (r *Reconciler) update() error {
-	if err := validateMachine(*r.machine, *r.providerSpec); err != nil {
+	if err := validateMachine(*r.machine); err != nil {
 		return fmt.Errorf("%v: failed validating machine provider spec: %v", r.machine.GetName(), err)
 	}
 
-	motask, err := r.session.GetTask(r.Context, r.providerStatus.TaskRef)
-	if err != nil {
-		if !isRetrieveMONotFound(r.providerStatus.TaskRef, err) {
+	if r.providerStatus.TaskRef != "" {
+		motask, err := r.session.GetTask(r.Context, r.providerStatus.TaskRef)
+		if err != nil {
+			if !isRetrieveMONotFound(r.providerStatus.TaskRef, err) {
+				return err
+			}
+		}
+		if taskIsFinished, err := taskIsFinished(motask); err != nil || !taskIsFinished {
+			if !taskIsFinished {
+				return fmt.Errorf("task %v has not finished", motask.Reference().Value)
+			}
 			return err
 		}
 	}
-	if taskIsFinished, err := taskIsFinished(motask); err != nil || !taskIsFinished {
-		if !taskIsFinished {
-			return fmt.Errorf("task %v has not finished", motask.Reference().Value)
-		}
-		return err
-	}
 
 	vmRef, err := findVM(r.machineScope)
 	if err != nil {
-		if !IsNotFound(err) {
+		if !isNotFound(err) {
 			return err
 		}
-		return errors.Wrap(err, "vm not found on update")
+		return fmt.Errorf("vm not found on update: %w", err)
 	}
 
 	vm := &virtualMachine{
@@ -116,22 +135,21 @@ func (r *Reconciler) update() error {
 		Ref:     vmRef,
 	}
 
-	// TODO: we won't always want to reconcile power state
-	//  but as per comment in clone() function, powering on right on creation might be problematic
-	if ok, err := vm.reconcilePowerState(); err != nil || !ok {
-		return err
+	if err := vm.reconcileTags(r.Context, r.session, r.machine); err != nil {
+		return fmt.Errorf("failed to reconcile tags: %w", err)
 	}
-	return r.reconcileMachineWithCloudState(vm)
+
+	return r.reconcileMachineWithCloudState(vm, r.providerStatus.TaskRef)
 }
 
 // exists returns true if machine exists.
 func (r *Reconciler) exists() (bool, error) {
-	if err := validateMachine(*r.machine, *r.providerSpec); err != nil {
+	if err := validateMachine(*r.machine); err != nil {
 		return false, fmt.Errorf("%v: failed validating machine provider spec: %v", r.machine.GetName(), err)
 	}
 
 	if _, err := findVM(r.machineScope); err != nil {
-		if !IsNotFound(err) {
+		if !isNotFound(err) {
 			return false, err
 		}
 		klog.Infof("%v: does not exist", r.machine.GetName())
@@ -142,22 +160,24 @@ func (r *Reconciler) exists() (bool, error) {
 }
 
 func (r *Reconciler) delete() error {
-	moTask, err := r.session.GetTask(r.Context, r.providerStatus.TaskRef)
-	if err != nil {
-		if !isRetrieveMONotFound(r.providerStatus.TaskRef, err) {
+	if r.providerStatus.TaskRef != "" {
+		moTask, err := r.session.GetTask(r.Context, r.providerStatus.TaskRef)
+		if err != nil {
+			if !isRetrieveMONotFound(r.providerStatus.TaskRef, err) {
+				return err
+			}
+		}
+		if taskIsFinished, err := taskIsFinished(moTask); err != nil || !taskIsFinished {
+			if !taskIsFinished {
+				return fmt.Errorf("task %v has not finished", moTask.Reference().Value)
+			}
 			return err
 		}
 	}
-	if taskIsFinished, err := taskIsFinished(moTask); err != nil || !taskIsFinished {
-		if !taskIsFinished {
-			return fmt.Errorf("task %v has not finished", moTask.Reference().Value)
-		}
-		return err
-	}
 
 	vmRef, err := findVM(r.machineScope)
 	if err != nil {
-		if !IsNotFound(err) {
+		if !isNotFound(err) {
 			return err
 		}
 		klog.Infof("%v: vm does not exist", r.machine.GetName())
@@ -178,13 +198,16 @@ func (r *Reconciler) delete() error {
 	if err != nil {
 		return fmt.Errorf("%v: failed to destroy vm: %v", r.machine.GetName(), err)
 	}
-	r.providerStatus.TaskRef = task.Reference().Value
+
+	if err := setProviderStatus(task.Reference().Value, conditionSuccess(), r.machineScope, vm); err != nil {
+		return fmt.Errorf("Failed to set provider status: %w", err)
+	}
 
 	return fmt.Errorf("destroying vm in progress, reconciling")
 }
 
 // reconcileMachineWithCloudState reconcile machineSpec and status with the latest cloud state
-func (r *Reconciler) reconcileMachineWithCloudState(vm *virtualMachine) error {
+func (r *Reconciler) reconcileMachineWithCloudState(vm *virtualMachine, taskRef string) error {
 	klog.V(3).Infof("%v: reconciling machine with cloud state", r.machine.GetName())
 	// TODO: reconcile task
 
@@ -197,8 +220,18 @@ func (r *Reconciler) reconcileMachineWithCloudState(vm *virtualMachine) error {
 	if err := r.reconcileProviderID(vm); err != nil {
 		return err
 	}
+
 	klog.V(3).Infof("%v: reconciling network", r.machine.GetName())
-	return r.reconcileNetwork(vm)
+	if err := r.reconcileNetwork(vm); err != nil {
+		return err
+	}
+
+	klog.V(3).Infof("%v: reconciling powerstate annotation", r.machine.GetName())
+	if err := r.reconcilePowerStateAnnontation(vm); err != nil {
+		return err
+	}
+
+	return setProviderStatus(taskRef, conditionSuccess(), r.machineScope, vm)
 }
 
 // reconcileRegionAndZoneLabels reconciles the labels on the Machine containing
@@ -266,7 +299,7 @@ func (r *Reconciler) reconcileNetwork(vm *virtualMachine) error {
 	//networks have IP addresses.
 	expectNetworkLen, currentNetworkLen := len(r.providerSpec.Network.Devices), len(currentNetworkStatusList)
 	if expectNetworkLen != currentNetworkLen {
-		return errors.Errorf("invalid network count: expected=%d current=%d", expectNetworkLen, currentNetworkLen)
+		return fmt.Errorf("invalid network count: expected=%d current=%d", expectNetworkLen, currentNetworkLen)
 	}
 
 	var ipAddrs []corev1.NodeAddress
@@ -279,12 +312,45 @@ func (r *Reconciler) reconcileNetwork(vm *virtualMachine) error {
 		}
 	}
 
+	// Using Name() if InventoryPath is empty will return empty name
+	// see: https://github.com/vmware/govmomi/blob/master/object/common.go#L66-L75
+	// Using ObjectName() as it will query from VirtualMachine properties
+
+	vmName, err := vm.Obj.ObjectName(vm.Context)
+	if err != nil {
+		return fmt.Errorf("error getting virtual machine name: %v", err)
+	}
+
+	ipAddrs = append(ipAddrs, corev1.NodeAddress{
+		Type:    corev1.NodeInternalDNS,
+		Address: vmName,
+	})
+
 	klog.V(3).Infof("%v: reconciling network: IP addresses: %v", r.machine.GetName(), ipAddrs)
 	r.machine.Status.Addresses = ipAddrs
 	return nil
 }
 
-func validateMachine(machine machinev1.Machine, providerSpec vsphereapi.VSphereMachineProviderSpec) error {
+func (r *Reconciler) reconcilePowerStateAnnontation(vm *virtualMachine) error {
+	if vm == nil {
+		return errors.New("provided VM is nil")
+	}
+
+	// This can return an error if machine is being deleted
+	powerState, err := vm.getPowerState()
+	if err != nil {
+		return err
+	}
+
+	if r.machine.Annotations == nil {
+		r.machine.Annotations = map[string]string{}
+	}
+	r.machine.Annotations[machinecontroller.MachineInstanceStateAnnotationName] = string(powerState)
+
+	return nil
+}
+
+func validateMachine(machine machinev1.Machine) error {
 	if machine.Labels[machinev1.MachineClusterIDLabel] == "" {
 		return machinecontroller.InvalidMachineConfiguration("%v: missing %q label", machine.GetName(), machinev1.MachineClusterIDLabel)
 	}
@@ -294,14 +360,20 @@ func validateMachine(machine machinev1.Machine, providerSpec vsphereapi.VSphereM
 
 func findVM(s *machineScope) (types.ManagedObjectReference, error) {
 	uuid := string(s.machine.UID)
-	objRef, err := s.GetSession().FindRefByInstanceUUID(s, uuid)
+
+	vm, err := s.GetSession().FindVM(s.Context, uuid, s.machine.Name)
 	if err != nil {
+		if isNotFound(err) {
+			return types.ManagedObjectReference{}, errNotFound{instanceUUID: true, uuid: uuid}
+		}
 		return types.ManagedObjectReference{}, err
 	}
-	if objRef == nil {
+
+	if vm == nil {
 		return types.ManagedObjectReference{}, errNotFound{instanceUUID: true, uuid: uuid}
 	}
-	return objRef.Reference(), nil
+
+	return vm.Reference(), nil
 }
 
 // errNotFound is returned by the findVM function when a VM is not found.
@@ -317,9 +389,9 @@ func (e errNotFound) Error() string {
 	return fmt.Sprintf("vm with bios uuid %s not found", e.uuid)
 }
 
-func IsNotFound(err error) bool {
+func isNotFound(err error) bool {
 	switch err.(type) {
-	case errNotFound, *errNotFound:
+	case errNotFound, *errNotFound, *find.NotFoundError:
 		return true
 	default:
 		return false
@@ -330,15 +402,49 @@ func isRetrieveMONotFound(taskRef string, err error) bool {
 	return err.Error() == fmt.Sprintf("ServerFaultCode: The object 'vim.Task:%v' has already been deleted or has not been completely created", taskRef)
 }
 
-func clone(s *machineScope) error {
+func clone(s *machineScope) (string, error) {
 	userData, err := s.GetUserData()
 	if err != nil {
-		return err
+		return "", err
 	}
 
-	vmTemplate, err := s.GetSession().FindVM(*s, s.providerSpec.Template)
+	vmTemplate, err := s.GetSession().FindVM(*s, "", s.providerSpec.Template)
 	if err != nil {
-		return err
+		return "", err
+	}
+
+	var snapshotRef *types.ManagedObjectReference
+
+	// If a linked clone is requested then a MoRef for a snapshot must be
+	// found with which to perform the linked clone.
+	// Empty clone mode is linked clone
+	if s.providerSpec.CloneMode == "" || s.providerSpec.CloneMode == vspherev1.LinkedClone {
+		if s.providerSpec.Snapshot == "" {
+			klog.V(3).Infof("%v: no snapshot name provided, getting snapshot using template", s.machine.GetName())
+			var vm mo.VirtualMachine
+			if err := vmTemplate.Properties(s.Context, vmTemplate.Reference(), []string{"snapshot"}, &vm); err != nil {
+				return "", fmt.Errorf("error getting snapshot information for template %s: %w", vmTemplate.Name(), err)
+			}
+
+			if vm.Snapshot != nil {
+				snapshotRef = vm.Snapshot.CurrentSnapshot
+			}
+		} else {
+			klog.V(3).Infof("%v: searching for snapshot by name %s", s.machine.GetName(), s.providerSpec.Snapshot)
+			var err error
+			snapshotRef, err = vmTemplate.FindSnapshot(s.Context, s.providerSpec.Snapshot)
+			if err != nil {
+				klog.V(3).Infof("%v: failed to find snapshot %s", s.machine.GetName(), s.providerSpec.Snapshot)
+			}
+		}
+	}
+
+	// The type of clone operation depends on whether or not there is a snapshot
+	// from which to do a linked clone.
+	diskMoveType := fullCloneDiskMoveType
+	if snapshotRef != nil {
+		// TODO: write clone mode to status
+		diskMoveType = linkCloneDiskMoveType
 	}
 
 	var folderPath, datastorePath, resourcepoolPath string
@@ -350,17 +456,26 @@ func clone(s *machineScope) error {
 
 	folder, err := s.GetSession().Finder.FolderOrDefault(s, folderPath)
 	if err != nil {
-		return errors.Wrapf(err, "unable to get folder for %q", folderPath)
+		multipleFoundMsg := "multiple folders found, specify one in config"
+		notFoundMsg := "folder not found, specify valid value"
+		defaultError := fmt.Errorf("unable to get folder for %q: %w", folderPath, err)
+		return "", handleVSphereError(multipleFoundMsg, notFoundMsg, defaultError, err)
 	}
 
 	datastore, err := s.GetSession().Finder.DatastoreOrDefault(s, datastorePath)
 	if err != nil {
-		return errors.Wrapf(err, "unable to get datastore for %q", datastorePath)
+		multipleFoundMsg := "multiple datastores found, specify one in config"
+		notFoundMsg := "datastore not found, specify valid value"
+		defaultError := fmt.Errorf("unable to get datastore for %q: %w", datastorePath, err)
+		return "", handleVSphereError(multipleFoundMsg, notFoundMsg, defaultError, err)
 	}
 
 	resourcepool, err := s.GetSession().Finder.ResourcePoolOrDefault(s, resourcepoolPath)
 	if err != nil {
-		return errors.Wrapf(err, "unable to get resource pool for %q", resourcepool)
+		multipleFoundMsg := "multiple resource pools found, specify one in config"
+		notFoundMsg := "resource pool not found, specify valid value"
+		defaultError := fmt.Errorf("unable to get resource pool for %q: %w", resourcepool, err)
+		return "", handleVSphereError(multipleFoundMsg, notFoundMsg, defaultError, err)
 	}
 
 	numCPUs := s.providerSpec.NumCPUs
@@ -378,28 +493,46 @@ func clone(s *machineScope) error {
 
 	devices, err := vmTemplate.Device(s.Context)
 	if err != nil {
-		return fmt.Errorf("error getting devices %v", err)
+		return "", fmt.Errorf("error getting devices %v", err)
+	}
+
+	// Create a new list of device specs for cloning the VM.
+	deviceSpecs := []types.BaseVirtualDeviceConfigSpec{}
+
+	// Only non-linked clones may expand the size of the template's disk.
+	if snapshotRef == nil {
+		diskSpec, err := getDiskSpec(s, devices)
+		if err != nil {
+			return "", fmt.Errorf("error getting disk spec for %q: %w", s.providerSpec.Snapshot, err)
+		}
+		deviceSpecs = append(deviceSpecs, diskSpec)
 	}
 
 	klog.V(3).Infof("Getting network devices")
 	networkDevices, err := getNetworkDevices(s, devices)
 	if err != nil {
-		return fmt.Errorf("error getting network specs: %v", err)
+		return "", fmt.Errorf("error getting network specs: %v", err)
 	}
 
-	var deviceSpecs []types.BaseVirtualDeviceConfigSpec
 	deviceSpecs = append(deviceSpecs, networkDevices...)
 
+	extraConfig := []types.BaseOptionValue{}
+
+	extraConfig = append(extraConfig, IgnitionConfig(userData)...)
+	extraConfig = append(extraConfig, &types.OptionValue{
+		Key:   GuestInfoHostname,
+		Value: s.machine.GetName(),
+	})
+
 	spec := types.VirtualMachineCloneSpec{
 		Config: &types.VirtualMachineConfigSpec{
 			Annotation: s.machine.GetName(),
 			// Assign the clone's InstanceUUID the value of the Kubernetes Machine
 			// object's UID. This allows lookup of the cloned VM prior to knowing
 			// the VM's UUID.
-			InstanceUuid: string(s.machine.UID),
-			Flags:        newVMFlagInfo(),
-			ExtraConfig:  IgnitionConfig(userData),
-			// TODO: set disk devices
+			InstanceUuid:      string(s.machine.UID),
+			Flags:             newVMFlagInfo(),
+			ExtraConfig:       extraConfig,
 			DeviceChange:      deviceSpecs,
 			NumCPUs:           numCPUs,
 			NumCoresPerSocket: numCoresPerSocket,
@@ -407,25 +540,36 @@ func clone(s *machineScope) error {
 		},
 		Location: types.VirtualMachineRelocateSpec{
 			Datastore:    types.NewReference(datastore.Reference()),
-			DiskMoveType: diskMoveType,
 			Folder:       types.NewReference(folder.Reference()),
 			Pool:         types.NewReference(resourcepool.Reference()),
+			DiskMoveType: diskMoveType,
 		},
-		// This is implicit, but making it explicit as it is important to not
-		// power the VM on before its virtual hardware is created and the MAC
-		// address(es) used to build and inject the VM with cloud-init metadata
-		// are generated.
-		PowerOn: false,
+		PowerOn:  true,
+		Snapshot: snapshotRef,
 	}
 
 	task, err := vmTemplate.Clone(s, folder, s.machine.GetName(), spec)
 	if err != nil {
-		return errors.Wrapf(err, "error triggering clone op for machine %v", s)
+		return "", fmt.Errorf("error triggering clone op for machine %v: %w", s, err)
 	}
 
-	s.providerStatus.TaskRef = task.Reference().Value
 	klog.V(3).Infof("%v: running task: %+v", s.machine.GetName(), s.providerStatus.TaskRef)
-	return nil
+	return task.Reference().Value, nil
+}
+
+func getDiskSpec(s *machineScope, devices object.VirtualDeviceList) (types.BaseVirtualDeviceConfigSpec, error) {
+	disks := devices.SelectByType((*types.VirtualDisk)(nil))
+	if len(disks) != 1 {
+		return nil, fmt.Errorf("invalid disk count: %d", len(disks))
+	}
+
+	disk := disks[0].(*types.VirtualDisk)
+	disk.CapacityInKB = int64(s.providerSpec.DiskGiB) * 1024 * 1024
+
+	return &types.VirtualDeviceConfigSpec{
+		Operation: types.VirtualDeviceConfigSpecOperationEdit,
+		Device:    disk,
+	}, nil
 }
 
 func getNetworkDevices(s *machineScope, devices object.VirtualDeviceList) ([]types.BaseVirtualDeviceConfigSpec, error) {
@@ -445,17 +589,17 @@ func getNetworkDevices(s *machineScope, devices object.VirtualDeviceList) ([]typ
 
 		ref, err := s.GetSession().Finder.Network(s.Context, netSpec.NetworkName)
 		if err != nil {
-			return nil, errors.Wrapf(err, "unable to find network %q", netSpec.NetworkName)
+			return nil, fmt.Errorf("unable to find network %q: %w", netSpec.NetworkName, err)
 		}
 
 		backing, err := ref.EthernetCardBackingInfo(s.Context)
 		if err != nil {
-			return nil, errors.Wrapf(err, "unable to create new ethernet card backing info for network %q", netSpec.NetworkName)
+			return nil, fmt.Errorf("unable to create new ethernet card backing info for network %q: %w", netSpec.NetworkName, err)
 		}
 
 		dev, err := object.EthernetCardTypes().CreateEthernetCard(ethCardType, backing)
 		if err != nil {
-			return nil, errors.Wrapf(err, "unable to create new ethernet card %q for network %q", ethCardType, netSpec.NetworkName)
+			return nil, fmt.Errorf("unable to create new ethernet card %q for network %q: %w", ethCardType, netSpec.NetworkName, err)
 		}
 
 		// Get the actual NIC object. This is safe to assert without a check
@@ -501,10 +645,50 @@ func taskIsFinished(task *mo.Task) (bool, error) {
 	case types.TaskInfoStateError:
 		return true, nil
 	default:
-		return false, errors.Errorf("task: %v, unknown state %v", task.Reference().Value, task.Info.State)
+		return false, fmt.Errorf("task: %v, unknown state %v", task.Reference().Value, task.Info.State)
 	}
 }
 
+func setProviderStatus(taskRef string, condition vspherev1.VSphereMachineProviderCondition, scope *machineScope, vm *virtualMachine) error {
+	klog.Infof("%s: Updating provider status", scope.machine.Name)
+
+	if vm != nil {
+		id := vm.Obj.UUID(scope.Context)
+		scope.providerStatus.InstanceID = &id
+
+		// This can return an error if machine is being deleted
+		powerState, err := vm.getPowerState()
+		if err != nil {
+			klog.V(3).Infof("%s: Failed to get power state during provider status update: %v", scope.machine.Name, err)
+		} else {
+			powerStateString := string(powerState)
+			scope.providerStatus.InstanceState = &powerStateString
+		}
+	}
+
+	if taskRef != "" {
+		scope.providerStatus.TaskRef = taskRef
+	}
+
+	scope.providerStatus.Conditions = setVSphereMachineProviderConditions(condition, scope.providerStatus.Conditions)
+
+	return nil
+}
+
+func handleVSphereError(multipleFoundMsg, notFoundMsg string, defaultError, vsphereError error) error {
+	var multipleFoundError *find.MultipleFoundError
+	if errors.As(vsphereError, &multipleFoundError) {
+		return machinecontroller.InvalidMachineConfiguration(multipleFoundMsg)
+	}
+
+	var notFoundError *find.NotFoundError
+	if errors.As(vsphereError, &notFoundError) {
+		return machinecontroller.InvalidMachineConfiguration(notFoundMsg)
+	}
+
+	return defaultError
+}
+
 type virtualMachine struct {
 	context.Context
 	Ref types.ManagedObjectReference
@@ -576,30 +760,6 @@ func (vm *virtualMachine) getRegionAndZone(c *rest.Client, regionLabel, zoneLabe
 	return result, nil
 }
 
-func (vm *virtualMachine) reconcilePowerState() (bool, error) {
-	powerState, err := vm.getPowerState()
-	if err != nil {
-		return false, err
-	}
-	switch powerState {
-	case types.VirtualMachinePowerStatePoweredOff:
-		klog.Infof("%v: powering on", vm.Obj.Reference().Value)
-		_, err := vm.powerOnVM()
-		if err != nil {
-			return false, errors.Wrapf(err, "failed to trigger power on op for vm %q", vm)
-		}
-		// TODO: store task in providerStatus/conditions?
-		klog.Infof("%v: requeue to wait for power on state", vm.Obj.Reference().Value)
-		return false, nil
-	case types.VirtualMachinePowerStatePoweredOn:
-		klog.Infof("%v: powered on", vm.Obj.Reference().Value)
-	default:
-		return false, errors.Errorf("unexpected power state %q for vm %q", powerState, vm)
-	}
-
-	return true, nil
-}
-
 func (vm *virtualMachine) powerOnVM() (string, error) {
 	task, err := vm.Obj.PowerOn(vm.Context)
 	if err != nil {
@@ -630,10 +790,57 @@ func (vm *virtualMachine) getPowerState() (types.VirtualMachinePowerState, error
 	case types.VirtualMachinePowerStateSuspended:
 		return types.VirtualMachinePowerStateSuspended, nil
 	default:
-		return "", errors.Errorf("unexpected power state %q for vm %v", powerState, vm)
+		return "", fmt.Errorf("unexpected power state %q for vm %v", powerState, vm)
 	}
 }
 
+// reconcileTags ensures that the required tags are present on the virtual machine, eg the Cluster ID
+// that is used by the installer on cluster deletion to ensure ther are no leaked resources.
+func (vm *virtualMachine) reconcileTags(ctx context.Context, session *session.Session, machine *machinev1.Machine) error {
+	if err := session.WithRestClient(vm.Context, func(c *rest.Client) error {
+		klog.Infof("%v: Reconciling attached tags", machine.GetName())
+
+		m := tags.NewManager(c)
+
+		clusterID := machine.Labels[machinev1.MachineClusterIDLabel]
+
+		attached, err := vm.checkAttachedTag(ctx, clusterID, m)
+		if err != nil {
+			return err
+		}
+
+		if !attached {
+			klog.Infof("%v: Attaching %s tag to vm", machine.GetName(), clusterID)
+			// the tag should already be created by installer
+			if err := m.AttachTag(ctx, clusterID, vm.Ref); err != nil {
+				return err
+			}
+		}
+
+		return nil
+	}); err != nil {
+		return err
+	}
+
+	return nil
+}
+
+// checkAttachedTag returns true if tag is already attached to a vm
+func (vm *virtualMachine) checkAttachedTag(ctx context.Context, tagName string, m *tags.Manager) (bool, error) {
+	tags, err := m.GetAttachedTags(ctx, vm.Ref)
+	if err != nil {
+		return false, err
+	}
+
+	for _, tag := range tags {
+		if tag.Name == tagName {
+			return true, nil
+		}
+	}
+
+	return false, nil
+}
+
 type NetworkStatus struct {
 	// Connected is a flag that indicates whether this network is currently
 	// connected to the VM.
@@ -658,7 +865,7 @@ func (vm *virtualMachine) getNetworkStatusList(client *vim25.Client) ([]NetworkS
 	}
 
 	if err := pc.RetrieveOne(vm.Context, vm.Ref, props, &obj); err != nil {
-		return nil, errors.Wrapf(err, "unable to fetch props %v for vm %v", props, vm.Ref)
+		return nil, fmt.Errorf("unable to fetch props %v for vm %v: %w", props, vm.Ref, err)
 	}
 	klog.V(3).Infof("Getting network status: object reference: %v", obj.Reference().Value)
 	if obj.Config == nil {
